# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "MCEq"
copyright = "2026, Anatoli Fedynitch, Stefan Fröse"
author = "Anatoli Fedynitch, Stefan Fröse"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
    "sphinx_copybutton",
    "sphinx_design",
    "nbsphinx",
    "sphinx_gallery.load_style",
    "matplotlib.sphinxext.plot_directive",
]


templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


source_suffix = ".rst"
master_doc = "index"


intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable", None),
    "crflux": ("https://crfluxmodels.readthedocs.io/en/latest", None),
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]

html_title = f"{project}"
htmlhelp_basename = project + "docs"

html_theme_options = {
    "navbar_links": [
        {"name": "Quickstart", "url": "quickstart/index", "internal": True},
        {"name": "Tutorial", "url": "tutorial", "internal": True},
        {"name": "Example Gallery", "url": "example-gallery", "internal": True},
        {"name": "v1.4", "url": "v14v13_diff", "internal": True},
        {"name": "Cite Us", "url": "citeus", "internal": True},
        {"name": "API Reference", "url": "api-reference/index", "internal": True},
    ],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/mceq-project/mceq",
            "icon": "fa-brands fa-github",
        },
    ],
    "announcement": "MCEq v1.4 is now available!",
}

html_sidebars = {
    "quickstart/index": [],  # disable sidebar on Quickstart
    "tutorial": [],
    "citeus": [],  # disable sidebar on Cite Us
    "example-gallery": [],  # disable sidebar on References
    "v12v11_diff": [],  # disable sidebar on References
    "v14v13_diff": [],  # disable sidebar on References
}

html_show_sourcelink = False
